@@ -1,6 +1,7 @@
 from __future__ import annotations
 
-import functools
+import collections
+import enum
 from typing import Literal
 
 import imgviz
@@ -33,8 +34,13 @@ CURSOR_GRAB = Qt.OpenHandCursor
 MOVE_SPEED = 5.0
 
 
+class CanvasMode(enum.Enum):
+    CREATE = enum.auto()
+    EDIT = enum.auto()
+
+
 class Canvas(QtWidgets.QWidget):
-    zoomRequest = QtCore.pyqtSignal(int, QPoint)
+    zoomRequest = QtCore.pyqtSignal(int, QPointF)
     scrollRequest = QtCore.pyqtSignal(int, int)
     newShape = QtCore.pyqtSignal()
     selectionChanged = QtCore.pyqtSignal(list)
@@ -44,7 +50,7 @@ class Canvas(QtWidgets.QWidget):
     mouseMoved = QtCore.pyqtSignal(QPointF)
     statusUpdated = QtCore.pyqtSignal(str)
 
-    CREATE, EDIT = 0, 1
+    mode: CanvasMode = CanvasMode.EDIT
 
     # polygon, rectangle, line, or point
     _createMode = "polygon"
@@ -55,11 +61,16 @@ class Canvas(QtWidgets.QWidget):
     prevMovePoint: QPointF
     offsets: tuple[QPointF, QPointF]
 
+    _dragging_start_pos: QPointF
+    _is_dragging: bool
+    _is_dragging_enabled: bool
+
     hVertex: int | None
     hShape: Shape | None
 
     _ai_model_name: str = "sam2:latest"
     _ai_model_cache: osam.types.Model | None = None
+    _ai_image_embedding_cache: collections.deque[tuple[str, osam.types.ImageEmbedding]]
 
     _cursor: QtCore.Qt.CursorShape
 
@@ -86,7 +97,6 @@ class Canvas(QtWidgets.QWidget):
         )
         super().__init__(*args, **kwargs)
         # Initialise local state.
-        self.mode = self.EDIT
         self.shapes = []
         self.shapesBackups = []
         self.current = None
@@ -103,6 +113,7 @@ class Canvas(QtWidgets.QWidget):
         self.offsets = QPointF(), QPointF()
         self.scale = 1.0
         self.pixmap = QtGui.QPixmap()
+        self._ai_image_embedding_cache = collections.deque(maxlen=3)
         self.visible = {}
         self._hideBackround = False
         self.hideBackround = False
@@ -117,6 +128,9 @@ class Canvas(QtWidgets.QWidget):
         self.hShapeIsSelected = False
         self._painter = QtGui.QPainter()
         self._cursor = CURSOR_DEFAULT
+        self._dragging_start_pos = QPointF()
+        self._is_dragging = False
+        self._is_dragging_enabled = False
         # Menus:
         # 0: right-click without selection and dragging of shapes
         # 1: right-click with selection and dragging of shapes
@@ -163,6 +177,28 @@ class Canvas(QtWidgets.QWidget):
         self._ai_model_cache = model_type()
         return self._ai_model_cache
 
+    def _get_ai_image_embedding(self) -> osam.types.ImageEmbedding:
+        qimage: QtGui.QImage = self.pixmap.toImage()
+
+        def pixmap_hash() -> int:
+            bits = qimage.constBits()
+            if bits is None:
+                return hash(None)
+            return hash(bits.asstring(qimage.sizeInBytes()))
+
+        cache_key: str = f"{self._ai_model_name}_{pixmap_hash()}"
+        key: str
+        image_embedding: osam.types.ImageEmbedding
+        for key, image_embedding in self._ai_image_embedding_cache:
+            if key == cache_key:
+                return image_embedding
+
+        image: np.ndarray = labelme.utils.img_qt_to_arr(img_qt=qimage)
+        image_embedding = self._get_ai_model().encode_image(image=imgviz.asrgb(image))
+        self._ai_image_embedding_cache.append((cache_key, image_embedding))
+        logger.debug("cached image embedding for key: {!r}", cache_key)
+        return image_embedding
+
     def storeShapes(self):
         shapesBackup = []
         for shape in self.shapes:
@@ -214,14 +250,14 @@ class Canvas(QtWidgets.QWidget):
         return self.visible.get(shape, True)
 
     def drawing(self):
-        return self.mode == self.CREATE
+        return self.mode == CanvasMode.CREATE
 
     def editing(self):
-        return self.mode == self.EDIT
+        return self.mode == CanvasMode.EDIT
 
     def setEditing(self, value=True):
-        self.mode = self.EDIT if value else self.CREATE
-        if self.mode == self.EDIT:
+        self.mode = CanvasMode.EDIT if value else CanvasMode.CREATE
+        if self.mode == CanvasMode.EDIT:
             # CREATE -> EDIT
             self.repaint()  # clear crosshair
         else:
@@ -258,7 +294,7 @@ class Canvas(QtWidgets.QWidget):
             messages.append(self.tr("Editing shapes"))
         if extra_messages:
             messages.extend(extra_messages)
-        self.statusUpdated.emit(" ー ".join(messages))
+        self.statusUpdated.emit(" • ".join(messages))
 
     def _get_create_mode_message(self) -> str:
         assert self.drawing()
@@ -308,6 +344,13 @@ class Canvas(QtWidgets.QWidget):
 
         is_shift_pressed = ev.modifiers() & Qt.ShiftModifier
 
+        if self._is_dragging:
+            self.overrideCursor(CURSOR_GRAB)
+            delta: QPointF = pos - self._dragging_start_pos
+            self.scrollRequest.emit(int(delta.x()), Qt.Horizontal)
+            self.scrollRequest.emit(int(delta.y()), Qt.Vertical)
+            return
+
         # Polygon drawing.
         if self.drawing():
             if self.createMode in ["ai_polygon", "ai_mask"]:
@@ -397,7 +440,9 @@ class Canvas(QtWidgets.QWidget):
         # - Highlight vertex
         # Update shape/vertex fill and tooltip value accordingly.
         status_messages: list[str] = []
-        for shape in reversed([s for s in self.shapes if self.isVisible(s)]):
+        for shape in ([self.hShape] if self.hShape else []) + [
+            s for s in reversed(self.shapes) if self.isVisible(s) and s != self.hShape
+        ]:
             # Look for a nearby vertex to highlight. If that fails,
             # check if we happen to be inside a shape.
             index = shape.nearestVertex(pos, self.epsilon)
@@ -563,6 +608,10 @@ class Canvas(QtWidgets.QWidget):
                 self.selectShapePoint(pos, multiple_selection_mode=group_mode)
                 self.repaint()
             self.prevPoint = pos
+        elif ev.button() == Qt.MiddleButton and self._is_dragging_enabled:
+            self.overrideCursor(CURSOR_GRAB)
+            self._dragging_start_pos = pos
+            self._is_dragging = True
         self._update_status()
 
     def mouseReleaseEvent(self, ev):
@@ -583,6 +632,9 @@ class Canvas(QtWidgets.QWidget):
                     self.selectionChanged.emit(
                         [x for x in self.selectedShapes if x != self.hShape]
                     )
+        elif ev.button() == Qt.MiddleButton:
+            self._is_dragging = False
+            self.restoreCursor()
 
         if self.movingShape and self.hShape:
             index = self.shapes.index(self.hShape)
@@ -828,7 +880,7 @@ class Canvas(QtWidgets.QWidget):
             )
             _update_shape_with_sam(
                 sam=self._get_ai_model(),
-                pixmap=self.pixmap,
+                image_embedding=self._get_ai_image_embedding(),
                 shape=drawing_shape,
                 createMode=self.createMode,
             )
@@ -841,6 +893,9 @@ class Canvas(QtWidgets.QWidget):
         """Convert from widget-logical coordinates to painter-logical ones."""
         return point / self.scale - self.offsetToCenter()
 
+    def enableDragging(self, enabled: bool):
+        self._is_dragging_enabled = enabled
+
     def offsetToCenter(self) -> QPointF:
         s = self.scale
         area = super().size()
@@ -859,7 +914,7 @@ class Canvas(QtWidgets.QWidget):
         if self.createMode in ["ai_polygon", "ai_mask"]:
             _update_shape_with_sam(
                 sam=self._get_ai_model(),
-                pixmap=self.pixmap,
+                image_embedding=self._get_ai_image_embedding(),
                 shape=self.current,
                 createMode=self.createMode,
             )
@@ -940,9 +995,15 @@ class Canvas(QtWidgets.QWidget):
         return self.minimumSizeHint()
 
     def minimumSizeHint(self):
-        if self.pixmap:
-            return self.scale * self.pixmap.size()
-        return super().minimumSizeHint()
+        if not self.pixmap:
+            return super().minimumSizeHint()
+
+        min_size = self.scale * self.pixmap.size()
+        if self._is_dragging_enabled:
+            # When drag buffer should be enabled, add a bit of buffer around the image
+            # This lets dragging the image around have a bit of give on the edges
+            min_size = 1.167 * min_size
+        return min_size
 
     def wheelEvent(self, ev: QtGui.QWheelEvent) -> None:
         mods: Qt.KeyboardModifiers = ev.modifiers()
@@ -950,7 +1011,7 @@ class Canvas(QtWidgets.QWidget):
         if Qt.ControlModifier == int(mods):
             # with Ctrl/Command key
             # zoom
-            self.zoomRequest.emit(delta.y(), ev.pos())
+            self.zoomRequest.emit(delta.y(), ev.posF())
         else:
             # scroll
             self.scrollRequest.emit(delta.x(), Qt.Horizontal)
@@ -1071,13 +1132,14 @@ class Canvas(QtWidgets.QWidget):
     def resetState(self):
         self.restoreCursor()
         self.pixmap = QtGui.QPixmap()
+        self.shapes = []
         self.shapesBackups = []
         self.update()
 
 
 def _update_shape_with_sam(
     sam: osam.types.Model,
-    pixmap: QtGui.QPixmap,
+    image_embedding: osam.types.ImageEmbedding,
     shape: Shape,
     createMode: Literal["ai_polygon", "ai_mask"],
 ) -> None:
@@ -1086,10 +1148,6 @@ def _update_shape_with_sam(
             f"createMode must be 'ai_polygon' or 'ai_mask', not {createMode}"
         )
 
-    image_embedding: osam.types.ImageEmbedding = _compute_image_embedding(
-        sam=sam, pixmap=pixmap
-    )
-
     response: osam.types.GenerateResponse = sam.generate(
         request=osam.types.GenerateRequest(
             model=sam.name,
@@ -1135,32 +1193,3 @@ def _update_shape_with_sam(
             points=[QPointF(point[0], point[1]) for point in points],
             point_labels=[1] * len(points),
         )
-
-
-def _compute_image_embedding(
-    sam: osam.types.Model, pixmap: QtGui.QPixmap
-) -> osam.types.ImageEmbedding:
-    return __compute_image_embedding(sam=sam, pixmap=_QPixmapForLruCache(pixmap))
-
-
-class _QPixmapForLruCache(QtGui.QPixmap):
-    def __hash__(self) -> int:
-        qimage: QtGui.QImage = self.toImage()
-        bits = qimage.constBits()
-        if bits is None:
-            return hash(None)
-        return hash(bits.asstring(qimage.sizeInBytes()))
-
-    def __eq__(self, other) -> bool:
-        if not isinstance(other, _QPixmapForLruCache):
-            return False
-        return self.__hash__() == other.__hash__()
-
-
-@functools.lru_cache(maxsize=3)
-def __compute_image_embedding(
-    sam: osam.types.Model, pixmap: _QPixmapForLruCache
-) -> osam.types.ImageEmbedding:
-    logger.debug("Computing image embeddings for model {!r}", sam.name)
-    image: np.ndarray = labelme.utils.img_qt_to_arr(pixmap.toImage())
-    return sam.encode_image(image=imgviz.asrgb(image))
